// Copyright 2018 The Rust Project Developers. See the COPYRIGHT
// file at the top-level directory of this distribution and at
// http://rust-lang.org/COPYRIGHT.
//
// Licensed under the Apache License, Version 2.0 <LICENSE-APACHE or
// http://www.apache.org/licenses/LICENSE-2.0> or the MIT license
// <LICENSE-MIT or http://opensource.org/licenses/MIT>, at your
// option. This file may not be copied, modified, or distributed
// except according to those terms.

use mem;
use sys::cloudabi::abi;
use time::Duration;

const NSEC_PER_SEC: abi::timestamp = 1_000_000_000;

#[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Debug, Hash)]
pub struct Instant {
    t: abi::timestamp,
}

pub fn checked_dur2intervals(dur: &Duration) -> Option<abi::timestamp> {
    dur.as_secs()
        .checked_mul(NSEC_PER_SEC)?
        .checked_add(dur.subsec_nanos() as abi::timestamp)
}

impl Instant {
    pub fn now() -> Instant {
        unsafe {
            let mut t = mem::uninitialized();
            let ret = abi::clock_time_get(abi::clockid::MONOTONIC, 0, &mut t);
            assert_eq!(ret, abi::errno::SUCCESS);
            Instant { t }
        }
    }

    pub fn sub_instant(&self, other: &Instant) -> Duration {
        let diff = self.t
            .checked_sub(other.t)
            .expect("second instant is later than self");
        Duration::new(diff / NSEC_PER_SEC, (diff % NSEC_PER_SEC) as u32)
    }

    pub fn checked_add_duration(&self, other: &Duration) -> Option<Instant> {
        Some(Instant {
            t: self.t.checked_add(checked_dur2intervals(other)?)?,
        })
    }

    pub fn checked_sub_duration(&self, other: &Duration) -> Option<Instant> {
        Some(Instant {
            t: self.t.checked_sub(checked_dur2intervals(other)?)?,
        })
    }
}

#[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord, Debug, Hash)]
pub struct SystemTime {
    t: abi::timestamp,
}

impl SystemTime {
    pub fn now() -> SystemTime {
        unsafe {
            let mut t = mem::uninitialized();
            let ret = abi::clock_time_get(abi::clockid::REALTIME, 0, &mut t);
            assert_eq!(ret, abi::errno::SUCCESS);
            SystemTime { t }
        }
    }

    pub fn sub_time(&self, other: &SystemTime) -> Result<Duration, Duration> {
        if self.t >= other.t {
            let diff = self.t - other.t;
            Ok(Duration::new(
                diff / NSEC_PER_SEC,
                (diff % NSEC_PER_SEC) as u32,
            ))
        } else {
            let diff = other.t - self.t;
            Err(Duration::new(
                diff / NSEC_PER_SEC,
                (diff % NSEC_PER_SEC) as u32,
            ))
        }
    }

    pub fn checked_add_duration(&self, other: &Duration) -> Option<SystemTime> {
        Some(SystemTime {
            t: self.t.checked_add(checked_dur2intervals(other)?)?,
        })
    }

    pub fn checked_sub_duration(&self, other: &Duration) -> Option<SystemTime> {
        Some(SystemTime {
            t: self.t.checked_sub(checked_dur2intervals(other)?)?,
        })
    }
}

pub const UNIX_EPOCH: SystemTime = SystemTime { t: 0 };
